@@ -4,8 +4,8 @@ from pyutil.assertutil import precondition, _assert
 
 from collections import deque
 
-# Note: when changing this class, you should un-comment all the lines that say
-# "assert self._assert_invariants()".
+# Note: when changing this class, you should set this to True. test_chain.py sets it to True.
+debug = False
 
 class StringChain(object):
     def __init__(self):
@@ -16,7 +16,7 @@ class StringChain(object):
 
     def append(self, s):
         """ Add s to the end of the chain. """
-        #assert self._assert_invariants()
+        #self._assert_invariants()
         if not s:
             return
 
@@ -27,11 +27,11 @@ class StringChain(object):
 
         self.d.append(s)
         self.len += len(s)
-        #assert self._assert_invariants()
+        #self._assert_invariants()
 
     def appendleft(self, s):
         """ Add s to the beginning of the chain. """
-        #assert self._assert_invariants()
+        #self._assert_invariants()
         if not s:
             return
 
@@ -42,32 +42,47 @@ class StringChain(object):
 
         self.d.appendleft(s)
         self.len += len(s)
-        #assert self._assert_invariants()
+        #self._assert_invariants()
 
-    def __str__(self):
+    def DISABLED__str__(self):
         """ Return the entire contents of this chain as a single
-        string. (Obviously this requires copying all of the bytes, so don't do
-        this unless you need to.) This has a side-effect of collecting all the
-        bytes in this StringChain object into a single string which is stored
-        in the first element of its internal deque. """
+        string. (Obviously this requires copying all of the bytes, so don't
+        do this unless you need to.) This has a side-effect of collecting all
+        the bytes in this StringChain object into a single string which is
+        stored in the first element of its internal deque."""
         self._collapse()
         if self.d:
             return self.d[0]
         else:
             return ''
 
-    def popleft_new_stringchain(self, bytes):
+    def popall(self):
+        """ Return the entire contents of this chain as a single string and
+        empty this chain. (Obviously this requires copying all of the bytes,
+        so don't do this unless you need to.)"""
+        res = self.__str__()
+        self.clear()
+        return res
+
+    def popall_new_stringchain(self):
+        """ Return the entire contents of this chain and empty this
+        chain."""
+        res = self.copy()
+        self.clear()
+        return res
+
+    def popleft_new_stringchain(self, numbytes):
         """ Remove some of the leading bytes of the chain and return them as a
         new StringChain object. (Use str() on it if you want the bytes in a
         string, or call popleft() instead of popleft_new_stringchain().) """
-        #assert self._assert_invariants()
-        if not bytes or not self.d:
+        #self._assert_invariants()
+        if not numbytes or not self.d:
             return self.__class__()
 
-        precondition(bytes >= 0, bytes)
+        precondition(numbytes > 0, numbytes)
 
         # We need to add at least this many bytes to the new StringChain.
-        bytesleft = bytes + self.ignored
+        bytesleft = numbytes + self.ignored
         n = self.__class__()
         n.ignored = self.ignored
 
@@ -91,23 +106,62 @@ class StringChain(object):
             self.ignored = 0
 
         # Either you got exactly how many you asked for, or you drained self entirely and you asked for more than you got.
-        #assert (n.len == bytes) or ((not self.d) and (bytes > self.len)), (n.len, bytes, len(self.d))
+        # if debug:
+        #     _assert((n.len == numbytes) or ((not self.d) and (numbytes > self.len)), (n.len, numbytes, len(self.d)))
 
-        #assert self._assert_invariants()
-        #assert n._assert_invariants()
+        # self._assert_invariants()
+        # n._assert_invariants()
         return n
 
-    def popleft(self, bytes):
+    def popleft_at_least_new_stringchain(self, numbytes):
+        """ Remove some of the leading bytes of the chain -- at least
+        `numbytes` of them -- and return them as a new StringChain
+        object. (Use str() on it if you want the bytes in a string, or call
+        popleft() instead of popleft_new_stringchain().)
+
+        This can be slightly more efficient than popleft_new_stringchain(),
+        since it doesn't have to slice a string in order to get exactly
+        `numbytes`."""
+        # self._assert_invariants()
+        if not numbytes or not self.d:
+            return self.__class__()
+
+        precondition(numbytes > 0, numbytes)
+
+        # We need to add at least this many bytes to the new StringChain.
+        bytesleft = numbytes + self.ignored
+        n = self.__class__()
+        n.ignored = self.ignored
+
+        while bytesleft > 0 and self.d:
+            s = self.d.popleft()
+            self.len -= (len(s) - self.ignored)
+            n.d.append(s)
+            n.len += (len(s)-self.ignored)
+            self.ignored = 0
+            bytesleft -= len(s)
+
+        self.ignored = 0
+
+        # Either you got at least how many you asked for, or you drained self entirely and you asked for more than you got.
+        # if debug:
+        #     _assert((n.len == numbytes) or ((not self.d) and (numbytes > self.len)), (n.len, numbytes, len(self.d)))
+
+        # self._assert_invariants()
+        # n._assert_invariants()
+        return n
+
+    def popleft(self, numbytes):
         """ Remove some of the leading bytes of the chain and return them as a
         string. """
-        #assert self._assert_invariants()
-        if not bytes or not self.d:
+        # self._assert_invariants()
+        if not numbytes or not self.d:
             return ''
 
-        precondition(bytes >= 0, bytes)
+        precondition(numbytes > 0, numbytes)
 
         # We need to add at least this many bytes to the result.
-        bytesleft = bytes
+        bytesleft = numbytes
         resstrs = []
 
         s = self.d.popleft()
@@ -135,21 +189,61 @@ class StringChain(object):
         resstr = ''.join(resstrs)
 
         # Either you got exactly how many you asked for, or you drained self entirely and you asked for more than you got.
-        #assert (len(resstr) == bytes) or ((not self.d) and (bytes > self.len)), (len(resstr), bytes, len(self.d), overrun)
+        # if debug:
+        #     _assert((len(resstr) == numbytes) or ((not self.d) and (numbytes > self.len)), (len(resstr), numbytes, len(self.d), overrun))
 
-        #assert self._assert_invariants()
+        # self._assert_invariants()
+
+        return resstr
+
+    def popleft_at_least(self, numbytes):
+        """ Remove some of the leading bytes of the chain -- at least
+        `numbytes` of them -- and return them as a string. This can can be
+        slightly more efficient than popleft(), since it doesn't have to
+        slice a string in order to get exactly `numbytes`."""
+        # self._assert_invariants()
+        if not numbytes or not self.d:
+            return ''
+
+        precondition(numbytes > 0, numbytes)
+
+        # We need to add at least this many bytes to the result.
+        bytesleft = numbytes
+        resstrs = []
+
+        s = self.d.popleft()
+        if self.ignored:
+            s = s[self.ignored:]
+            self.ignored = 0
+        self.len -= len(s)
+        resstrs.append(s)
+        bytesleft -= len(s)
+
+        while bytesleft > 0 and self.d:
+            s = self.d.popleft()
+            self.len -= len(s)
+            resstrs.append(s)
+            bytesleft -= len(s)
+
+        resstr = ''.join(resstrs)
+
+        # Either you got at least how many you asked for, or you drained self entirely and you asked for more than you got.
+        # if debug:
+        #     _assert((len(resstr) >= numbytes) or ((not self.d) and (numbytes > self.len)), (len(resstr), numbytes, len(self.d)))
+
+        # self._assert_invariants()
 
         return resstr
 
     def __len__(self):
-        #assert self._assert_invariants()
+        # self._assert_invariants()
         return self.len
 
-    def trim(self, bytes):
+    def trim(self, numbytes):
         """ Trim off some of the leading bytes. """
-        #assert self._assert_invariants()
-        self.ignored += bytes
-        self.len -= bytes
+        # self._assert_invariants()
+        self.ignored += numbytes
+        self.len -= numbytes
         while self.d and self.ignored >= len(self.d[0]):
             s = self.d.popleft()
             self.ignored -= len(s)
@@ -157,16 +251,16 @@ class StringChain(object):
             self.len = 0
         if not self.d:
             self.ignored = 0
-        #assert self._assert_invariants()
+        # self._assert_invariants()
 
     def clear(self):
         """ Empty it out. """
-        #assert self._assert_invariants()
+        # self._assert_invariants()
         self.d.clear()
         self.ignored = 0
         self.tailignored = 0
         self.len = 0
-        #assert self._assert_invariants()
+        # self._assert_invariants()
 
     def copy(self):
         n = self.__class__()
@@ -174,10 +268,12 @@ class StringChain(object):
         n.tailignored = self.tailignored
         n.len = self.len
         n.d = copy.copy(self.d)
-        #assert n._assert_invariants()
+        # n._assert_invariants()
         return n
 
     def _assert_invariants(self):
+        if not debug:
+            return True
         _assert(self.ignored >= 0, self.ignored)
         _assert(self.tailignored >= 0, self.tailignored)
         _assert(self.len >= 0, self.len)
@@ -191,7 +287,7 @@ class StringChain(object):
         """ Concatenate all of the strings into one string and make that string
         be the only element of the chain. (Obviously this requires copying all
         of the bytes, so don't do this unless you need to.) """
-        #assert self._assert_invariants()
+        # self._assert_invariants()
         # First trim off any leading ignored bytes.
         if self.ignored:
             self.d[0] = self.d[0][self.ignored:]
@@ -204,4 +300,4 @@ class StringChain(object):
             newstr = ''.join(self.d)
             self.d.clear()
             self.d.append(newstr)
-        #assert self._assert_invariants()
+        # self._assert_invariants()
